/*********************************************************************/
/* queues.h                                                          */
/* typedefs, macros and inline functs for various queues management  */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#ifndef NEBLO_QUEUES_H
#define NEBLO_QUEUES_H


#include <neblo_.h>

#undef DEBUG_Q                /* dealing with Queues (of many kinds) */


/*
 * queue of announced (expected) friendship requests.
 */

typedef struct expd_reqst
{
  struct in_addr foreign_addr;   /* IP addr of requestor                  */
  u_int16_t foreign_port;        /* listen port of requestor              */
  unsigned int pred_succ;        /* 1 if expected new predecessor,        */
                                 /* 2 if expected new successor,          */
                                 /* 0 otherwise.                          */
  struct expd_reqst *next;
  RSA *pubkey;                   /* pubkey of requestor */
  struct timeval timestamp;      /* time of insert, for the garbage coll. */
} expd_reqst_t;

void init_expd_reqst_q ( void );

RSA  *search_expd_reqst ( struct in_addr, u_int16_t, unsigned int, pthread_t );

int   insert_expd_reqst
        ( struct in_addr, u_int16_t, unsigned int, RSA *, pthread_t );

void *clean_expd_reqsts ( void * );


/*
 * queue of pending read requests:
 * request of kind "read" issued to a remote peer but not yet answered
 */

#define I_MAX_P_READ 16384      /* initial room for pending "read" reqsts */
#define MAX_P_READ 65536        /* max no. of pending "read" reqsts */

typedef struct pend_read_req
{
  pthread_mutex_t lock;         /* lock on entire struct */
  unsigned int in_use;
  unsigned int done;            /* significant on originator peer only */
                                /* 1 iff request served (data available) */
                                /* -1 iff got "data not found" answers only */
                                /* 0 iff no answer arrived yet */
  neblo_rhandle handle;         /* for now this is = ptr to the whole struct */
  peer_t *from;                 /* req came from this peer, so response */
                                /* will be eventually backwarded to it. */
                                /* NULL iff req comes from local peer */
  neblo_rhandle from_rhandle;   /* id. of request at the "from" peer */
                                /* R_COMPLETE iff req comes from local peer */
  neblo_port port;              /* read through this NEBLO port of recipient */
  neblo_addr addr;              /* dest. of the read req in the 64-bit space */
  neblo_addr real_addr;         /* read at this addr, in the 64-bit space */
  void *data;                   /* incoming data will be stored at this addr.*/
                                /* NULL iff req did not originate locally */
  unsigned int maxlen;          /* max data size for this read req */
  pthread_mutex_t wait_sem;     /* semaphore associated to condition below */
  pthread_cond_t wait_cond;     /* condition; will unblock when "read" done */
  struct timeval timestamp;     /* time of insertion, for the garbage coll. */
} pend_read_t;

void  init_pend_read_q ( void );

int   insert_pend_read
        ( neblo_port, neblo_addr, void *, unsigned int, peer_t *,
        neblo_rhandle, neblo_rhandle * );

void  remove_expd_reqst_from_successor ( pthread_t );

void *clean_pend_reads ( void * );

void  clean_pend_reads_peer ( peer_t * );

extern pend_read_t       *pend_read;
extern unsigned int       pend_read_head;
extern unsigned int       pend_read_tail;
extern unsigned int       pend_read_max;
extern pthread_mutex_t    p_read_tail_lock;

/* some of the routines for this queue are so simple that should be inlined */

/*
 * search by "rhandle".  "rhandle", as returned by previous call to
 * insert_read_reqst(), is ptr to an entry into the queue (casted to
 * neblo_rhandle), so direct access, no actual search
 */
static inline pend_read_t *
search_read_reqst
( neblo_rhandle h )
{

  pend_read_t *pr;


  pr = (pend_read_t *)h;
  if (pr != NULL)
    if (pr->handle == h && pr->in_use)
      return pr;
  return NULL;

}

/* cancel read request "pr".  NOTE: caller must have raised pr->lock */
static inline void
cancel_read_reqst
(pend_read_t *pr)
{
  if (pr != NULL)
    pr->in_use = 0;
  return;
}


/*
 * queue of read/write/data msgs that could not be forwarded yet, and thus
 * must be retried
 */

#define FIRST_ATTMPT 0
#define RETRY 1
#define LAST_CHANCE 2

typedef struct retry_reqst
{
  rwd_t kind;                     /* READ, WRITE, DATA, REFR */
  reqst_t *reqst;                 /* the request to be retried */
  unsigned int pending;           /* 1 if still to be served, 0 otherwise */
  struct retry_reqst *next;
  struct timeval timestamp;       /* time of insertion, for the garbg coll */
} retry_reqst_t;

void  init_retry_queue ( void );

int   insert_retry_queue ( rwd_t, reqst_t *, pthread_t );

void *clean_retry_queue ( void * );


#endif
